#! /usr/local/bin/scm \ %0 %*
- !#
;;; "nk" create, manage, and query spectra database
;;; Copyright (C) 2003, 2006, 2008, 2010 Aubrey Jaffer

;; This program is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.
;;
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program; if not, write to the Free Software
;; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

(require 'filename)
(require 'scanf)
(require 'printf)
(require 'line-i/o)
(require 'string-case)
(require 'string-search)
(require 'metric-units)
(require 'rwb-isam)
(require 'databases)
(require 'database-interpolate)
(require 'eps-graph)
(require-if 'compiling 'read-command)
(require-if 'compiling 'directory)

(if (not (defined? real-log10))
    (define real-log10 $log10))

(define (string-starts-with-ci? str prefix)
  (define len (string-length prefix))
  (if (< (string-length str) len)
      #f
      (string-ci=? (substring str 0 len) prefix)))

(define (string-ends-with-ci? str suffix)
  (define pos (- (string-length str) (string-length suffix)))
  (if (< pos 0)
      #f
      (string-ci=? (substring str pos (string-length str)) suffix)))

(define (nk.usage)
  (display "\

Use the nk program to create, manage, and query a refractive-index
spectra database.

The nk program can read Sopra format files.  These files have a \".nk\"
suffix.  A bundle of spectra for over 275 materials is available from:
http://www.sopra-sa.com/index2.php?goto=dl&rub=4 (2010-06-14).  Some
metal spectra included with FreeSnell also have a \".nk\" suffix, but
with a different format.

The zip file http://www.sspectra.com/files/misc/win/SOPRA.EXE
contains \"README.TXT\" identifying the Sopra files.

http://luxpop.com/RefractiveIndexList.html has 274 refractive-index
files, mostly X-ray spectra.
"
	   (current-error-port))
  (nki.usage))

(define (nki.usage)
  (display "\

The optional argument [-F path] sets the path to the database file to
be accessed or created.  If it is not given, then the environment
variable \"NK_DATABASE_PATH\" names the database file if defined,
defaulting to \"nk.rwb\" in the current directory.  The name for each
spectrum in the database is taken from its filename sans the \".nk\"
or \"ir.nk\" suffix.

\
Usage: nk [-F path]
\
  Starts the nk shell.  Type nk commands without the nk; ^Z to exit.
\
Usage: nk [-F path] --add FILEGLOB.nk ...
\
  Add spectra matching FILEGLOB to database.  Wildcards must be in the
  leaf directory only.
\
Usage: nk [-F path] --new FILEGLOB.nk ...
\
  Add new or replace spectra matching FILEGLOB to database.  Wildcards
  must be in the leaf directory only.
\
Usage: nk [-F path] --del 'GLOB' ...
\
  Delete spectra of names matching GLOB from database.
\
Usage: nk [-F path] --plot 'GLOB' ...
Usage: nk [-F path] --lin-lin 'GLOB' ...
Usage: nk [-F path] --log-log 'GLOB' ...
Usage: nk [-F path] --log-lin 'GLOB' ...
Usage: nk [-F path] --lin-log 'GLOB' ...
Usage: nk [-F path] --lin-eV 'GLOB' ...
Usage: nk [-F path] --log-eV 'GLOB' ...
Usage: nk [-F path] --lin-/cm 'GLOB' ...
Usage: nk [-F path] --log-/cm 'GLOB' ...
\
  Create encapsulated-PostScript (and view with Ghostview) spectra plots
  from database of names matching GLOB.  The first symbol controls the
  y-axis.
\
Usage: nk [-F path] --list 'GLOB' ...
\
  List names of spectra matching GLOB in database.
\
Usage: nk [-F path] --range 'GLOB' ...
\
  List names and data ranges of spectra matching GLOB in database.
\
Usage: nk [-F path] --desc 'GLOB' ...
\
  List names and descriptions of spectra matching GLOB in database.
\
Usage: nk [-F path] --annotate sopra.txt
\
  Add chemical-symbol and description annotations to database.
\
Usage: nk [-F path] NAME NUMBER.UNIT ...
\
  Prints NAME's n+k*i values at each NUMBER.UNIT from database.  The
  NUMBER can be fixed or floating point; the UNIT either \"eV\"
  (electron-volt) or \"m\" (meter) with an optional metric prefix or
  \"cm^-1\" (wave-number).
\
Usage: nk --help
\
  Prints this help and exits.
\

"
	   (current-error-port))
  (force-output (current-error-port))
  #f)

(define (nk.script args)
  (cond ((not (<= 1 (length args)))
	 (nki))
	((equal? "-F" (car args))
	 (cond ((null? (cdr args)) (nk.usage))
	       (else (set! *nkdb-filename* (cadr args))
		     (nk.script (cddr args)))))
	((eqv? #\- (string-ref (car args) 0))
	 (case (string->symbol (strip-leading-dashes (car args)))
	   ((lin-eV) (apply plot-nk-spectra #f 'eV (cdr args)))
	   ((log-eV) (apply plot-nk-spectra #t 'eV (cdr args)))
	   ((lin-/cm) (apply plot-nk-spectra #f '/cm (cdr args)))
	   ((log-/cm) (apply plot-nk-spectra #t '/cm (cdr args)))
	   ((log-log) (apply plot-nk-spectra #t 'log (cdr args)))
	   ((lin-log) (apply plot-nk-spectra #f 'log (cdr args)))
	   ((log-lin) (apply plot-nk-spectra #t 'lin (cdr args)))
	   ((lin-lin
	     plot)    (apply plot-nk-spectra #f 'lin (cdr args)))
	   ((list) (apply list-nk-spectra (cdr args)))
	   ((range) (apply list-spectra-range (cdr args)))
	   ((desc) (apply list-spectra-desc (cdr args)))
	   ((new) (apply add-nk-spectra #t (cdr args)))
	   ((add) (apply add-nk-spectra #f (cdr args)))
	   ((del) (apply del-nk-spectra (cdr args)))
	   ((annotate) (apply add-annotations (cdr args)))
	   ((help) (nk.usage) #t)
	   #+IGNORE
	   ((version)
	    (call-with-input-file
		(in-vicinity (freesnell-vicinity) "version.txi")
	      (lambda (iprt)
		(read iprt)
		(read iprt)
		(write (read iprt) (current-error-port))
		(force-output (current-error-port))
		#t)))
	   (else (nk.usage))))
	(else (apply nk-values args))))

(define (nki.prompt-read-command)
  (display "nk ")
  (force-output)
  (read-command))

(define (nki.welcome)
  (display
   (case (software-type)
     ((ms-dos windows atarist os/2)
      "Complete the nk command lines; ? for help; ^Z to quit.")
     (else
      "Complete the nk command lines; ? for help; ^D to quit.")))
  (newline))

(define (nki)				; interactive
  (require 'read-command)
  (nki.welcome)
  (do ((args (nki.prompt-read-command) (nki.prompt-read-command)))
      ((or (eof-object? args))
       (exit #t))
    (cond ((null? args))
	  ((equal? "?" (car args))
	   (nk.usage)
	   (nki.welcome))
	  (else
	   (nk.script args)))))

(define (strip-leading-dashes str)
  (define len (string-length str))
  (do ((idx 0 (+ 1 idx)))
      ((or (>= idx len)
	   (not (eqv? #\- (string-ref str idx))))
       (substring str idx len))))
				 ; Energy	     E =         h*c/L
(define eV<->L			 ; Photon wavelength L =         h*c/E
  (let ((c 299792458)		 ; speed of light    c =           m/s
	(h 6.62606876e-34)	 ; Plank's constant  h =           J.s
	(J/eV 1.602176462e-19))
    (lambda (eV) (/ (* h c) eV J/eV))))

(define (recip-cm x) (/ .01 x))

(define (readit port)
  (let ((num (read port)))
    (cond ((eqv? #\, (peek-char port))
	   (read-char port)))
    num))

(define (string-whitespace? str)
  (do ((idx (+ -1 (string-length str)) (+ -1 idx)))
      ((or (negative? idx)
	   (not (char-whitespace? (string-ref str idx))))
       (negative? idx))))

;;This is the filename of the nk database.
(define *nkdb-filename* (or (getenv "NK_DATABASE_PATH") "nk.rwb"))

;;Database is global.
(define *nkdb* #f)

;;Error routine.
(define (dno name) (slib:warn name 'database 'not 'opened *nkdb-filename*) #f)

(define (create-nk-database)
  (set! *nkdb* (create-database *nkdb-filename* 'rwb-isam))
  (cond (*nkdb*
	 (define-domains *nkdb*
	   '(wavelength #f real? r32 "m")
	   '(refractive-index #f complex? c32 ""))
	 (define-tables *nkdb*
	   '(*desc:spectrum*
	     *columns*
	     *columns*
	     ((1 #t wavelength #f wavelength)
	      (2 #f refractive-index #f refractive-index)))
	   '(materials
	     ((symbol symbol))
	     ((name string)
	      (description string))
	     ()))
;;; Don't remember why this reopen is here.
	 (close-database *nkdb*)
	 (open-database! *nkdb-filename* 'rwb-isam))
	(else (slib:warn 'create-nk-database 'failed *nkdb-filename*)
	      #f)))

(define (nk:open-database! filename type create?)
  (cond ((file-exists? filename)
	 (let ((nkdb (open-database! filename type)))
	   (cond ((not nkdb)
		  (describe-file-lock *nkdb-filename*)
		  (slib:warn 'could 'not 'open *nkdb-filename* 'for 'writing)))
	   nkdb))
	(create?
	 (let ((nkdb (create-nk-database)))
	   (cond ((not nkdb)
		  (slib:warn 'could 'not 'create *nkdb-filename*)))
	   nkdb))
	(else
	 (slib:warn 'could 'not 'find *nkdb-filename*)
	 #f)))

;;http://www.sopra-sa.com/more/database.asp
;;
;; Database n&k
;;
;; SOPRA provides with their instruments one of the largest database of
;; optical indices available in the world (278 materials ).  We have now
;; decided to give this database for free to the scientific community.
;; It can be download very easily here.  Nevertheless the inverse can
;; also be true and if you send us your own optical indices with the
;; allowance to add them in the database their can be useful to other
;; people.  This database comes with a help file which provide some more
;; details about these data and also their origin.
;;
;; File Format:
;;
;; The different NK files are in ASCII format (*.NK files).  The first
;; number indicates the spectral unit (1 = eV, 2 = m, 3 = cm-1, 4 = nm).
;; The second and the third ones indicate the spectral range and the
;; fourth one the number of intervals.  The N and K values are then
;; stored in the order N,K,N,K... One example is reported below:
;;
;; 1, 0.6, 5.9, 53
;; 3.4471, 0.
;; 3.4595, 0.
;; ...
;; 1.1328, 3.0447
;; 1.0831, 2.9823
;;
;;     * NK database with 278 materials.
;;       http://www.sopra-sa.com/fichiers/Database.zip

(define (load-spectrum.nk path spectrum-table te?)
  (define (micronize wl) (* 1e-6 wl))
  (define (nanoize wl) (* 1e-9 wl))
  (define (Angstromize wl) (* 1e-10 wl))
  (define spectrum:insert (spectrum-table (if te? 'row:update 'row:insert)))
  (call-with-open-ports
   (open-file path (_tracked open_read))
   (lambda (port)
     (define pow #f)			; previous row
     (define pine #f)			; previous line
     (define (report class . args)
       (printf "%s: %d: %s: " path (+ -1 (port-line port)) class)
       (for-each (lambda (x)
		   (if (number? x) (printf "%g" x) (write x))
		   (display " "))
		 args)
       (display "\n"))
     ;; reader for formats: wl|wn|ev n k ...
     (define (reader converter second)
       (printf "%s: %s%s\n" path (if (eq? 'a first) "" first) second)
       (call-with-current-continuation
	(lambda (escape)
	  (do ((line (read-line port) (read-line port)))
	      ((eof-object? line))
	    (let ((row (scanf-read-list " %g %g %g" line)))
	      (define (doins)
		(set! pow row)
		(set! pine line)
		(spectrum:insert
		 (list (converter (car row))
		       (make-rectangular (cadr row) (max 0 (caddr row))))))
	      (if (or (negative? (cadr row))
		      (negative? (caddr row)))
		  (report 'warning 'negative? line))
	      (cond
	       ((and pow (> (car pow) (car row)))
		(report 'error 'abscissa-not-increasing (car pow) '> (car row))
		(printf "%s\n%s\n" pine line)
		(escape #f))
	       ((and pow (= (car pow) (car row)))
		(report 'warning 'repeated-abscissa (car pow))
		(printf "%s\n%s\n" pine line))
	       ((and pow (<= (* 9 (car pow)) (car row)))
		(report 'error 'abscissa-gap-too-large (car pow) '<< (car row))
		(printf "%s\n%s\n" pine line)
		(escape #f))
	       ((and pow (< (* 3 (car pow)) (car row)))
		(report 'warning 'abscissa-gap (car pow) '< (car row))
		(printf "%s\n%s\n" pine line)
		(doins))
	       (else
		(doins)))))))
       (list "" (let ((sdx (substring? "; " second)))
		  (if sdx
		      (substring second (+ 2 sdx) (string-length second))
		      second))))
     (define first (peek-char port))
     (define head1 #f)
     (define cite #f)
     (define (semi-blank? str)
       (and (string-starts-with-ci? str ";")
	    (do ((idx (+ -1 (string-length str)) (+ -1 idx)))
		((or (< idx 1)
		     (not (char-whitespace? (string-ref str idx))))
		 (< idx 1)))))
     (define (luxpop-header)
       (set! head1 (read-line port))
       (cond
	((string-starts-with-ci? head1 ";  Optical constants for ")
	 (set! head1 (substring head1
				(string-length ";  Optical constants for ")
				(string-length head1)))
	 (set! first (read-line port))
	 (cond ((semi-blank? first)
		#t)
	       ((string-starts-with-ci? first "; ")
		(set! head1 (string-append
			     head1
			     (substring first 2 (string-length first))))
		(set! first (read-line port))
		(semi-blank? first))
	       (else #f)))
	(else #f)))
     (cond
      ((and
	(eqv? #\; first)
	(luxpop-header)
	(set! cite (scanf-read-list ";   taken from:
;
;   %[^;]
;   %[^;]
;   
;    Lambda (A)            n            %s
;-----------------------------------------
"
				    port))
	(list? cite)
	(= 3 (length cite))
	(equal? '"k" (caddr cite)))
       (set! first 'A))
      (else (set! first (readit port))))
     (case first
       ;; luxpop.com format
       ((A) (reader Angstromize head1))
       ;; FreeSnell formats for CRC data
       ((eV) (reader eV<->L (read-line port)))
       ((/cm) (reader recip-cm (read-line port)))
       ((um) (reader micronize (read-line port)))
       ((nm) (reader nanoize (read-line port)))
       (else
	;; Sopra formats
	(let ((format (case first ((1) 'eV) ((2) 'um) ((3) '/cm) ((4) nm)
			    (else (report 'warning 'bad 'first 'line? first)
				  #f))))
	  (and
	   format
	   (let* ((conv-lm (case format
			     ((eV) identity)
			     ((um) micronize)
			     ((/cm) identity)
			     ((nm) nanoize)))
		  (conv-dt (case format
			     ((eV) eV<->L)
			     ((um) identity)
			     ((/cm) recip-cm)
			     ((nm) identity)))
		  (start   (readit port))
		  (end     (readit port))
		  (count-1 (readit port)))
	     (define wlinc (conv-lm (/ (- end start) count-1)))
	     (define cnt 0)
	     (define eline (read-line port))
	     (printf "%s: %.3g.%s :: %.3g.%s\n" path start format end format)
	     (if (not (string-whitespace? eline)) (write-line eline))
	     (do ((wl (conv-lm start) (+ wl wlinc))
		  (val (readit port) (readit port)))
		 ((not (number? val))
		  (if (not (eqv? cnt (+ 1 count-1)))
		      (report 'warning 'Inserted cnt 'of (+ 1 count-1) 'values))
		  '("" ""))
	       (let ((imag (readit port)))
		 (read-line port)
		 (if (negative? imag)
		     (report 'warning 'negative 'k imag))
		 (if (negative? val)
		     (report 'warning 'negative 'n val))
		 (cond ((not (number? imag))
			(report 'warning 'strange 'imag imag))
		       (else (spectrum:insert
			      (list (conv-dt wl)
				    (make-rectangular val (max 0 imag))))
			     (set! cnt (+ 1 cnt))))))))))))))

;;@body
;;@3 must be an open relational database, @2 a symbol, and the string
;;@1 must name an existing file with spectral values.  @0 creates a
;;table @2 in @3 and enters the associations found in @1 into it.
(define (file->spectrum! file table-name te?)
  (cond ((not (file-exists? file))	; logic error
	 (slib:error 'load-spectrum! 'file-exists? file)))
  (if (not te?)
      (define-tables *nkdb*
	`(,table-name
	  *desc:spectrum*
	  *desc:spectrum*
	  ())
	`(catalog
	  ((name symbol))
	  ()
	  ((,table-name)))))
  (let ((table ((*nkdb* 'open-table) table-name #t)))
    (cond (table
	   (let ((desc (load-spectrum.nk file table te?))
		 (materials (open-table! *nkdb* 'materials)))
	     (define mat-row
	       (or ((materials 'row:retrieve) table-name)
		   '(#f "" "")))
	     (and desc
		  materials
		  ((materials 'row:update)
		   (list table-name
			 (if (equal? "" (cadr mat-row))
			     (car desc)
			     (cadr mat-row))
			 (if (equal? "" (caddr mat-row))
			     (cadr desc)
			     (caddr mat-row))))))))))

(define (add-annotations file)
  (define materials #f)
  (cond
   ((not (file-exists? file))		; logic error
    (slib:warn 'add-annotations 'file-exists? file)
    #f)
   ((begin (set! *nkdb* (open-database! *nkdb-filename* 'rwb-isam))
	   (not *nkdb*))
    (dno 'add-annotations))
   ((begin (set! materials (open-table! *nkdb* 'materials))
	   (not materials))
    (close-database *nkdb*)
    (slib:warn 'add-annotations 'could-not-open-table 'materials)
    #f)
   (else
    (call-with-input-file file
      (lambda (port)
	(define mru (materials 'row:update))
	(define mrr (materials 'row:retrieve))
	(do ((line (read-line port) (read-line port)))
	    ((eof-object? line)
	     (and (close-database *nkdb*)
		  (set! *nkdb* #f)
		  #t))
	  (let ((lst (scanf-read-list "%s %s %d %d %[^\n]" line)))
	    (cond
	     ((eof-object? lst))
	     ((not (list? lst)) (print '>>>> 'add-annotations lst))
	     ((= 5 (length lst))
	      (let ((symbol (rootname->symbol (string-downcase (car lst)))))
		(define mat-row (mrr symbol))
		(cond (mat-row
		       (mru
			(list symbol
			      (cadr lst)
			      (car (cddddr lst)))))
		      (else (print 'not-found lst)))))
	     (else (printf "%s\n" line))))))))))

;;; Sopra/sicrir.nk and Sopra/sio2ir.nk extend the range of
;;; sicr.nk (crystalline Si) and sio2.nk respectively.
(define (add-nk-spectra new? . files)
  (define path-syms '())
  (for-each (lambda (path)
	      (define dir (pathname->vicinity path))
	      (define fnd? #f)
	      (cond
	       ((file-exists? path)
		(set! path-syms
		      (cons
		       (cons path
			     (substring path
					(string-length dir)
					(+ -3 (string-length path))))
		       path-syms))
		(set! fnd? #t))
	       (else
		(let ()
		  (require 'directory)
		  (directory*-for-each
		   (lambda (fname)
		     (set! path-syms
			   (cons
			    (cons
			     (string-append dir fname)
			     (substring fname 0 (+ -3 (string-length fname))))
			    path-syms))
		     (set! fnd? #t))
		   path))))
	      (cond ((not fnd?)
		     (slib:warn 'add-nk-spectra 'can 'not 'find path)
		     (set! files #f))))
	    files)
  (cond
   ((not files) (slib:warn 'add-nk-spectra 'not 'done) #f)
   (else
    (set! *nkdb* (nk:open-database! *nkdb-filename* 'rwb-isam #t))
    (cond (*nkdb*
	   (let ((table-exists? (*nkdb* 'table-exists?)))
	     (for-each
	      (lambda (path-sym)
		(define nam (rootname->symbol (cdr path-sym)))
		(let ((te? (table-exists? nam)))
		  (if (and new? te?)
		      ((*nkdb* 'delete-table) nam))
		  (file->spectrum! (car path-sym) nam (and te? (not new?)))))
	      path-syms)
	     (and (close-database *nkdb*)
		  (set! *nkdb* #f)
		  #t)))
	  (else
	   (dno 'add-nk-spectra))))))

(define (rootname->symbol nam)
  (define len (string-length nam))
  (set! nam
	(cond ((and (> len 3) (string-ends-with-ci? nam "ir"))
	       ;; Sopra sicrir.nk sio2ir.nk
	       (print nam '==> (substring nam 0 (+ -2 len))))
	      ((string-ends-with-ci? nam "_palik")
	       ;; luxpop.com
	       (substring nam 0 (+ -6 len)))
	      ((string-ends-with-ci? nam "-a")
	       ;; -a is extraordinary ray
	       (print nam '==> (replace-suffix nam "-a" "-e")))
	      ((string-ends-with-ci? nam "-b")
	       ;; -b is ordinary ray
	       (print nam '==> (substring nam 0 (+ -2 len))))
	      (else nam)))
  (set! len (string-length nam))
  (string-ci->symbol
   (cond ((string-starts-with-ci? nam "sicr")
	  ;; Sopra sicr* is really Si
	  (print nam '==> (string-append "si" (substring nam 4 len))))
	 ((string-ci=? "diam" nam)
	  (print nam '==> "d-C"))
	 ;; Use broader Palik data (which is incompatible with al2o3.nk).
	 ((string-ci=? "al2o3p" nam)
	  (print nam '==> "al2o3"))
	 ((and (> len 4) (string-ends-with-ci? nam "cub"))
	  ;; Sopra znscub.nk
	  (print nam '==> (substring nam 0 (+ -3 len))))
	 ((string-starts-with-ci? nam "ir-")
	  ;; luxpop.com ir-ZnSe_palik.nk ir-ZnS_palik.nk
	  (print nam '==> (substring nam 3 len)))
	 (else nam))))

(define (find-matching *nkdb* globs)
  (define cat (open-table *nkdb* 'materials))
  (define names '())
  (let ((catget* (and cat (cat 'get* 'symbol))))
    (and catget*
	 (if (null? globs)
	     (set! names (catget*))
	     (for-each
	      (lambda (glob)
		(define match? (filename:match-ci?? glob))
		(let ((finds (catget* (lambda (key)
					(match? (symbol->string key))))))
		  (cond ((null? finds) (slib:warn glob 'not 'found))
			(else (set! names (append finds names))))))
	      globs))
	 names)))

(define (list-spectra-desc . globs)
  (define any-error? #f)
  (set! *nkdb* (open-database *nkdb-filename* 'rwb-isam))
  (cond (*nkdb*
	 (let ((cat (open-table *nkdb* 'materials)))
	   (define for-each-row (and cat (cat 'for-each-row)))
	   (define (proc row) (apply printf "%10s  %s %s\n" row))
	   (cond ((not for-each-row) #f)
		 ((null? globs)
		  (for-each-row proc)
		  #t)
		 (else
		  (for-each (lambda (glob)
			      (define match? (filename:match-ci?? glob))
			      (define found? #f)
			      (for-each-row
			       (lambda (row)
				 (set! found? #t)
				 (proc row))
			       (lambda (key) (match? (symbol->string key))))
			      (cond ((not found?)
				     (set! any-error? #t)
				     (slib:warn glob 'not 'found))))
			    globs)
		  (not any-error?))))
	 (and (close-database *nkdb*)
	      (set! *nkdb* #f)
	      #t))
	(else (dno 'list-spectra-desc))))

(define (del-nk-spectra . substances)
  (set! *nkdb* (nk:open-database! *nkdb-filename* 'rwb-isam #f))
  (cond (*nkdb*
	 (let ((delete-table (*nkdb* 'delete-table))
	       (cat (open-table! *nkdb* 'catalog))
	       (mat (open-table! *nkdb* 'materials)))
	   (define catrem (and cat (cat 'row:remove)))
	   (define matrem (and mat (mat 'row:remove)))
	   (for-each (lambda (substance)
		       (printf "removing %s\n" substance)
		       (delete-table substance)
		       (catrem substance)
		       (matrem substance))
		     (find-matching *nkdb* substances)))
	 (and (close-database *nkdb*)
	      (set! *nkdb* #f)
	      #t))
	(else (dno 'del-nk-spectra))))

(define (array<-complex-table table log-y? x-scl)
  (define rows '())
  (define warp-y (if log-y? real-log10 identity))
  (define warp-x (case x-scl
		   ((log) real-log10)
		   ((lin) identity)
		   ((/cm) recip-cm)
		   ((eV) eV<->L)
		   (else identity)))
  ((table 'for-each-row)
   (lambda (row)
     (set! rows (cons (list (warp-y (imag-part (cadr row)))
			    (warp-y (real-part (cadr row)))
			    (warp-x (car row)))
		      rows))))
  (let ((nra (make-array (A:floR64b)
			 (length rows)
			 (length (car rows)))))
    (do ((lns rows (cdr lns))
	 (idx (+ -1 (length rows)) (+ -1 idx)))
	((null? lns) nra)
      (do ((kdx (+ -1 (length (car rows))) (+ -1 kdx))
	   (lst (car lns) (cdr lst)))
	  ((null? lst))
	(array-set! nra (car lst) idx kdx)))))
;;(/ (imag-part (cadr row)) (real-part (cadr row)))

(define (graph-spectrum log-y? x-scl table-name)
  (define table (open-table *nkdb* table-name))
  (define tabra #f)
  (set! tabra (array<-complex-table table log-y? x-scl))
  (create-postscript-graph
   (sprintf #f "/tmp/%s.eps" table-name) '(600 400)
   (set-margin-templates "+200.0" "+200.0") ; for Aluminum
   (in-graphic-context (set-color 100) (fill-rect (whole-page)))
   (whole-page)
   (setup-plot (map (case x-scl
		      ((log) real-log10)
		      ((lin) identity)
		      ((/cm) recip-cm)
		      ((eV) eV<->L)
		      (else identity))
		    (append ((table 'isam-next) -inf.0)
			    ((table 'isam-prev) +inf.0)))
	       ;;(column-range tabra 0)
	       (combine-ranges (column-range tabra 1)
			       (column-range tabra 2)))
   (let ((row (((open-table *nkdb* 'materials) 'row:retrieve)
	       table-name)))
     (title-top (if (equal? "" (cadr row))
		    table-name
		    (sprintf #f "%s (%s)" table-name (cadr row)))
		(caddr row)))
   (outline-rect plotrect)
   (in-graphic-context (set-linewidth 0) (grid-verticals) (grid-horizontals))
   (rule-vertical leftedge
		  (if log-y? "log refractive-index n" "refractive-index n")
		  10)
   (rule-horizontal bottomedge
		    (case x-scl
		      ((log) "log(wavelength/m)")
		      ((lin) "wavelength in m")
		      ((/cm) "waves/cm")
		      ((eV) "eV")
		      (else "??"))
		    10)
   (plot-column tabra 0 1 'line)
   (plot-column tabra 0 1 'plus)
   (set-color 'seagreen)
   (rule-vertical rightedge
		  (if log-y? "log extinction-coefficient k" "extinction-coefficient k")
		  -10)
   (plot-column tabra 0 2 'line)
   (plot-column tabra 0 2 'cross))
  (zero? (system (case (software-type)
		   ((ms-dos windows atarist os/2)
		    (sprintf #f
			     "%s \"\\\\tmp\\\\%s.eps\""
			     "\\Progra~1\\Ghostgum\\gsview\\gsview32"
			     table-name))
		   (else
		    (sprintf #f
			     "%s \"/tmp/%s.eps\""
			     "gv"
			     table-name))))))

(define (plot-nk-spectra log-y? x-scl . substances)
  (set! *nkdb* (open-database *nkdb-filename* 'rwb-isam))
  (cond (*nkdb*
	 (for-each
	  (lambda (name)
	    (graph-spectrum log-y? x-scl name))
	  (find-matching *nkdb* substances))
	 (and (close-database *nkdb*)
	      (set! *nkdb* #f)
	      #t))
	(else (dno 'plot-nk-spectra))))

(define (nk-lookup substance itp low high wv)
  (define ok? #t)
  (define (try un)
    (cond
     ((or (not (list? un)) (not (eqv? 2 (length un)))) #f)
     (else
      (if (equal? "/cm" (cadr un)) (set! un (list (car un) "cm^-1")))
      (cond
       ((positive? (si:conversion-factor "cm^-1" (cadr un)))
	(let ((/cm (* (car un)
		      (si:conversion-factor "cm^-1" (cadr un)))))
	  (define w (recip-cm /cm))
	  (cond ((<= low w high)
		 (printf "%s: %.4g @ (%.3Km) (%.3g.eV) %g.cm^-1\n"
			 substance (itp w) w (eV<->L w) /cm))
		(else
		 (set! ok? #f)
		 (printf ">>>> %s: wavenumber %g not in range %g to %g.\n"
			 substance /cm (recip-cm high) (recip-cm low))))))
       ((positive? (si:conversion-factor "eV" (cadr un)))
	(let ((eV (* (car un)
		     (si:conversion-factor "eV" (cadr un)))))
	  (define w (eV<->L eV))
	  (cond ((<= low w high)
		 (printf "%s: %.4g @ (%.3Km) %.3g.eV (%g.cm^-1)\n"
			 substance (itp w) w eV (recip-cm w)))
		(else
		 (set! ok? #f)
		 (printf ">>>> %s: %.3g.eV not in range %.3g to %.3g.\n"
			 substance ev (eV<->L high) (eV<->L low))))))
       ((positive? (si:conversion-factor "m" (cadr un)))
	(let ((w (* (car un)
		    (si:conversion-factor "m" (cadr un)))))
	  (cond ((<= low w high)
		 (printf "%s: %.4g @ %.3Km (%.3g.eV) (%g.cm^-1)\n"
			 substance (itp w) w (eV<->L w) (recip-cm w)))
		(else
		 (set! ok? #f)
		 (printf
		  ">>>> %s: wavelength %.3Km not in range %.3Km to %.3Km.\n"
		  substance w low high)))
	  ))
       (else #f)))))
  (cond ((or (try (scanf-read-list "%g.%s" wv))
	     (try (scanf-read-list "%d.%s" wv))
	     (try (scanf-read-list "%g%s" wv)))
	 ok?)
	(else (printf "Unrecognized numeric.unit expression: '%s'\n" wv)
	      #f)))

(define (nk-values substance . wvs)
  (define ok? #t)
  (if (string? substance) (set! substance (string-ci->symbol substance)))
  (set! *nkdb* (open-database *nkdb-filename* 'rwb-isam))
  (cond (*nkdb*
	 (let ((table (open-table *nkdb* substance)))
	   (cond (table
		  (let ((itp (interpolate-from-table table 2))
			(low ((table 'isam-next) -inf.0))
			(high ((table 'isam-prev) +inf.0)))
		    (if high (set! high (car high)))
		    (if low (set! low (car low)))
		    (for-each (lambda (wv)
				(or (nk-lookup substance itp low high wv)
				    (set! ok? #f)))
			      wvs)))))
	 (and (close-database *nkdb*)
	      (set! *nkdb* #f)
	      ok?))
	(else (dno 'nk-values))))

(define (list-nk-spectra . globs)
  (define col 0)
  (define res 0)
  (set! *nkdb* (open-database *nkdb-filename* 'rwb-isam))
  (cond (*nkdb*
	 (for-each (lambda (name)
		     (define len (string-length (symbol->string name)))
		     (cond ((>= (+ col len res) 72)
			    (printf "\n%s" name)
			    (set! col (+ len)))
			   (else
			    (printf "%*s" (+ res len) name)
			    (set! col (+ col len res))))
		     (set! res (modulo (- len) 12)))
		   (find-matching *nkdb* globs))
	 (newline)
	 (and (close-database *nkdb*)
	      (set! *nkdb* #f)
	      #t))
	(else (dno 'list-nk-spectra))))

(define (list-spectra-range . globs)
  (define (pad-to goal used)
    (if (< used goal) (printf "%*s" (- goal used) "")))
  (set! *nkdb* (open-database *nkdb-filename* 'rwb-isam))
  (cond (*nkdb*
	 (for-each (lambda (name)
		     (let ((table (open-table *nkdb* name)))
		       (define low ((table 'isam-next) -inf.0))
		       (define high ((table 'isam-prev) +inf.0))
		       (cond
			((and low high)
			 (set! low (car low))
			 (set! high (car high))
			 (pad-to 10 (printf "%s" name))
			 (printf "%8s" (sprintf #f "%.3Km" low))
			 (pad-to 14 (printf " to %.3Km;" high))
			 (printf "%8s" (sprintf #f "%.3KeV" (eV<->L high)))
			 (pad-to 13 (printf " to %.3KeV;" (eV<->L low)))
			 (printf "%9s" (sprintf #f "%.0f/cm" (/ 0.01 high)))
			 (printf " to %.0f/cm\n" (/ 0.01 low))))))
		   (find-matching *nkdb* globs))
	 (newline)
	 (and (close-database *nkdb*)
	      (set! *nkdb* #f)
	      #t))
	(else (dno 'list-spectra-range))))

;;; procedure for "nk.bat".
(define (nk) (exit (nk.script (list-tail *argv* *optind*))))

;;(trace-all "/home/jaffer/slib/dbutil.scm") (untrace mdbm:report)

(if *script* (nk))
